using System;
using System.Collections;
using System.Data;

public class Attribute
{
	ArrayList mValues;
	String mName;
	object mLabel;

	public Attribute(String name, String[] values)
	{
		mName = name;
		mValues = new ArrayList(values);
		mValues.Sort();
	}

	public Attribute(object Label)
	{
		mLabel = Label;
		mName = String.Empty;
		mValues = null;
	}

	public String AttributeName
	{
		get
		{
			return mName;
		}
	}

		
	public String[] values
	{
		get
		{
			if (mValues != null)
				return (String[])mValues.ToArray(typeof(String));
			else
				return null;
		}
	}

	public bool isValidValue(String value)
	{
		return indexValue(value) >= 0;
	}

	public int indexValue(String value)
	{
		if (mValues != null)
			return mValues.BinarySearch(value);
		else
			return -1;
	}

	public override String ToString()
	{
		if (mName != String.Empty)
		{
			return mName;
		}
		else
		{
			return mLabel.ToString();
		}
	}
}


public class TreeNode
{
	private ArrayList mChilds = null;
	private Attribute mAttribute;

	public TreeNode(Attribute attribute)
	{
		if (attribute.values != null)
		{
			mChilds = new ArrayList(attribute.values.Length);
			for (int i = 0; i < attribute.values.Length; i++)
				mChilds.Add(null);
		}
		else
		{
			mChilds = new ArrayList(1);
			mChilds.Add(null);
		}
		mAttribute = attribute;
	}

	public void AddTreeNode(TreeNode treeNode, String ValueName)
	{
		int index = mAttribute.indexValue(ValueName);
		mChilds[index] = treeNode;
	}

	public int totalChilds
	{
		get
		{
			return mChilds.Count;
		}
	}


	public TreeNode getChild(int index)
	{
		return (TreeNode)mChilds[index];
	}

	public Attribute attribute
	{
		get
		{
			return mAttribute;
		}
	}

	public TreeNode getChildByBranchName(String branchName)
	{
		int index = mAttribute.indexValue(branchName);
		return (TreeNode)mChilds[index];
	}
}


public class DecisionTreeID3
{
	private DataTable mSamples;
	private int mTotalPositives = 0;
	private int mTotal = 0;
	private String mTargetAttribute = "result";
	private double mEntropySet = 0.0;

	private int countTotalPositives(DataTable samples)
	{
		int result = 0;
			foreach (DataRow aRow in samples.Rows)
		{
			if ((bool)aRow[mTargetAttribute] == true)
				result++;
		}
			return result;
	}

	private double calcEntropy(int positives, int negatives)
	{
		int total = positives + negatives;
		double ratioPositive = (double)positives/total;
		double ratioNegative = (double)negatives/total;
			if (ratioPositive != 0)
			ratioPositive = -(ratioPositive) * System.Math.Log(ratioPositive, 2);
		if (ratioNegative != 0)
			ratioNegative = - (ratioNegative) * System.Math.Log(ratioNegative, 2);
			double result =  ratioPositive + ratioNegative;
			return result;
	}

	private void getValuesToAttribute(DataTable samples, Attribute attribute, String value, out int positives, out int negatives)
	{
		positives = 0;
		negatives = 0;
		foreach (DataRow aRow in samples.Rows)
		{
			if (  ((String)aRow[attribute.AttributeName] == value) )
				if ( (bool)aRow[mTargetAttribute] == true) 
					positives++;
				else
					negatives++;
		}		
	}


	private double gain(DataTable samples, Attribute attribute)
	{
		String[] values = attribute.values;
		double sum = 0.0;
		for (int i = 0; i < values.Length; i++)
		{
			int positives, negatives;
			
			positives = negatives = 0;
			
			getValuesToAttribute(samples, attribute, values[i], out positives, out negatives);
			
				double entropy = calcEntropy(positives, negatives);				
			sum += -(double)(positives + negatives)/mTotal * entropy;
		}
		return mEntropySet + sum;
	}

	private Attribute getBestAttribute(DataTable samples, Attribute[] attributes)
	{
		double maxGain = 0.0;
		Attribute result = null;

		foreach (Attribute attribute in attributes)
		{
			double aux = gain(samples, attribute);
			if (aux > maxGain)
			{
				maxGain = aux;
				result = attribute;
			}
		}
		return result;
	}

	private bool allSamplesPositives(DataTable samples, String targetAttribute)
	{			
		foreach (DataRow row in samples.Rows)
		{
			if ( (bool)row[targetAttribute] == false)
				return false;
		}

		return true;
	}


	private bool allSamplesNegatives(DataTable samples, String targetAttribute)
	{
		foreach (DataRow row in samples.Rows)
		{
			if ( (bool)row[targetAttribute] == true)
				return false;
		}

		return true;			
	}

	private ArrayList getDistinctValues(DataTable samples, String targetAttribute)
	{
		ArrayList distinctValues = new ArrayList(samples.Rows.Count);
		foreach(DataRow row in samples.Rows)
		{
			if (distinctValues.IndexOf(row[targetAttribute]) == -1)
				distinctValues.Add(row[targetAttribute]);
		}

		return distinctValues;
	}


	private object getMostCommonValue(DataTable samples, String targetAttribute)
	{
		ArrayList distinctValues = getDistinctValues(samples, targetAttribute);
		int[] count = new int[distinctValues.Count];
		foreach(DataRow row in samples.Rows)
		{
			int index = distinctValues.IndexOf(row[targetAttribute]);
			count[index]++;
		}
		
		int MaxIndex = 0;
		int MaxCount = 0;

		for (int i = 0; i < count.Length; i++)
		{
			if (count[i] > MaxCount)
			{
				MaxCount = count[i];
				MaxIndex = i;
			}
		}

		return distinctValues[MaxIndex];
	}


	private TreeNode internalMountTree(DataTable samples, String targetAttribute, Attribute[] attributes)
	{
		if (allSamplesPositives(samples, targetAttribute) == true)
			return new TreeNode(new Attribute(true));
		
		if (allSamplesNegatives(samples, targetAttribute) == true)
			return new TreeNode(new Attribute(false));

		if (attributes.Length == 0)
			return new TreeNode(new Attribute(getMostCommonValue(samples, targetAttribute)));			
		
		mTotal = samples.Rows.Count;
		mTargetAttribute = targetAttribute;
		mTotalPositives = countTotalPositives(samples);
		mEntropySet = calcEntropy(mTotalPositives, mTotal - mTotalPositives);
		
		Attribute bestAttribute = getBestAttribute(samples, attributes); 
		TreeNode root = new TreeNode(bestAttribute);
			
		DataTable aSample = samples.Clone();			
			
		foreach(String value in bestAttribute.values)
		{				
		
			aSample.Rows.Clear();

			DataRow[] rows = samples.Select(bestAttribute.AttributeName + " = " + "'"  + value + "'");
		
			foreach(DataRow row in rows)
			{					
				aSample.Rows.Add(row.ItemArray);
			}				
				
			ArrayList aAttributes = new ArrayList(attributes.Length - 1);
			for(int i = 0; i < attributes.Length; i++)
			{
				if (attributes[i].AttributeName != bestAttribute.AttributeName)
					aAttributes.Add(attributes[i]);
			}

			if (aSample.Rows.Count == 0)
			{
				return new TreeNode(new Attribute(getMostCommonValue(aSample, targetAttribute)));
			}
			else
			{				
				DecisionTreeID3 dc3 = new DecisionTreeID3();
				TreeNode ChildNode =  dc3.mountTree(aSample, targetAttribute, (Attribute[])aAttributes.ToArray(typeof(Attribute)));
				root.AddTreeNode(ChildNode, value);
			}
		}
		return root;
	}


	public TreeNode mountTree(DataTable samples, String targetAttribute, Attribute[] attributes)
	{
		mSamples = samples;
		return internalMountTree(mSamples, targetAttribute, attributes);
	}
	
	public String printNode(TreeNode root, String tabs, String p)
	{
		String printout;
		printout = tabs + '|' + root.attribute + "|\n";
		
		if (root.attribute.values != null)
		{
			for (int i = 0; i < root.attribute.values.Length; i++)
			{
				printout += tabs + "\t" + "<" + root.attribute.values[i] + ">";
				TreeNode childNode = root.getChildByBranchName(root.attribute.values[i]);
				printout += printNode(childNode, "\t" + tabs, printout);
			}
		}
		return printout;
	}
	
	
	public TreeNode createDataTable()
	{
		Attribute a1 = new Attribute("num_minerals", new string[] {"Low", "High"});
		Attribute a2 = new Attribute("num_gas", new string[] {"Low", "High"});
		Attribute a3 = new Attribute("access", new string[] {"Yes", "No"});
		Attribute a4 = new Attribute("distance", new string[] {"Close", "Far"});

		Attribute[] attributes = new Attribute[] {a1, a2, a3, a4};
		
		
		DataTable result = new DataTable("test");
		DataColumn column = result.Columns.Add("num_minerals");
		column.DataType = typeof(String);
		column = result.Columns.Add("num_gas");
		column.DataType = typeof(String);
		column = result.Columns.Add("access");
		column.DataType = typeof(String);
		column = result.Columns.Add("distance");
		column.DataType = typeof(String);
		column = result.Columns.Add("strategy");
		column.DataType = typeof(bool);
	
		//minerals, gas, access, distance, priority defence?
		result.Rows.Add(new object[] {"Low","High","No","Far",true});
		result.Rows.Add(new object[] {"Low","High","Yes","Far",true});
		result.Rows.Add(new object[] {"Low","Low","No","Far",false});
		result.Rows.Add(new object[] {"Low","Low","Yes","Far",false});
		result.Rows.Add(new object[] {"High","High","No","Far",true});
		result.Rows.Add(new object[] {"High","High","Yes","Far",true});
		result.Rows.Add(new object[] {"High","Low","No","Far",true});
		result.Rows.Add(new object[] {"High","Low","Yes","Far",true});
		result.Rows.Add(new object[] {"Low","High","No","Close",true});
		result.Rows.Add(new object[] {"Low","High","Yes","Close",false});
		result.Rows.Add(new object[] {"Low","Low","No","Close",false});
		result.Rows.Add(new object[] {"Low","Low","Yes","Close",false});
		result.Rows.Add(new object[] {"High","High","No","Close",true});
		result.Rows.Add(new object[] {"High","High","Yes","Close",true});
		result.Rows.Add(new object[] {"High","Low","No","Close",true});
		result.Rows.Add(new object[] {"High","Low","Yes","Close",false});
		
		
		TreeNode root = mountTree(result, "strategy", attributes);
		return root;
	}

	
}



